from unittest import TestCase

from sekg.text.util import CodeTextPreprocessor


class TestCodeTextPreprocessor(TestCase):
    def test_clean_html_text(self):
        html_text = """
<div class="block"><p>
 The <tt>Marshaller</tt> class is responsible for governing the process
 of serializing Java content trees back into XML data.  It provides the basic
 marshalling methods:

 </p><p>
 <i>Assume the following setup code for all following code fragments:</i>
 </p><blockquote>
    <pre>       JAXBContext jc = JAXBContext.newInstance( "com.acme.foo" );
       Unmarshaller u = jc.createUnmarshaller();
       Object element = u.unmarshal( new File( "foo.xml" ) );
       Marshaller m = jc.createMarshaller();
    </pre>
 </blockquote>

 <p>
 Marshalling to a File:
 </p><blockquote>
    <pre>       OutputStream os = new FileOutputStream( "nosferatu.xml" );
       m.marshal( element, os );
    </pre>
 </blockquote>

 <p>
 Marshalling to a SAX ContentHandler:
 </p><blockquote>
    <pre>       // assume MyContentHandler instanceof ContentHandler
       m.marshal( element, new MyContentHandler() );
    </pre>
 </blockquote>

 <p>
 Marshalling to a DOM Node:
 </p><blockquote>
    <pre>       DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
       dbf.setNamespaceAware(true);
       DocumentBuilder db = dbf.newDocumentBuilder();
       Document doc = db.newDocument();

       m.marshal( element, doc );
    </pre>
 </blockquote>

 <p>
 Marshalling to a java.io.OutputStream:
 </p><blockquote>
    <pre>       m.marshal( element, System.out );
    </pre>
 </blockquote>

 <p>
 Marshalling to a java.io.Writer:
 </p><blockquote>
    <pre>       m.marshal( element, new PrintWriter( System.out ) );
    </pre>
 </blockquote>

 <p>
 Marshalling to a javax.xml.transform.SAXResult:
 </p><blockquote>
    <pre>       // assume MyContentHandler instanceof ContentHandler
       SAXResult result = new SAXResult( new MyContentHandler() );

       m.marshal( element, result );
    </pre>
 </blockquote>

 <p>
 Marshalling to a javax.xml.transform.DOMResult:
 </p><blockquote>
    <pre>       DOMResult result = new DOMResult();

       m.marshal( element, result );
    </pre>
 </blockquote>

 <p>
 Marshalling to a javax.xml.transform.StreamResult:
 </p><blockquote>
    <pre>       StreamResult result = new StreamResult( System.out );

       m.marshal( element, result );
    </pre>
 </blockquote>

 <p>
 Marshalling to a javax.xml.stream.XMLStreamWriter:
 </p><blockquote>
    <pre>       XMLStreamWriter xmlStreamWriter =
           XMLOutputFactory.newInstance().createXMLStreamWriter( ... );

       m.marshal( element, xmlStreamWriter );
    </pre>
 </blockquote>

 <p>
 Marshalling to a javax.xml.stream.XMLEventWriter:
 </p><blockquote>
    <pre>       XMLEventWriter xmlEventWriter =
           XMLOutputFactory.newInstance().createXMLEventWriter( ... );

       m.marshal( element, xmlEventWriter );
    </pre>
 </blockquote>

 <p>
 <a name="elementMarshalling"></a>
 <b>Marshalling content tree rooted by a JAXB element</b><br>
 </p><blockquote>
 The first parameter of the overloaded
 <tt>Marshaller.marshal(java.lang.Object, ...)</tt> methods must be a
 JAXB element as computed by
 <a href="../../../javax/xml/bind/JAXBIntrospector.html#isElement(java.lang.Object)"><code>JAXBIntrospector.isElement(java.lang.Object)</code></a>;
 otherwise, a <tt>Marshaller.marshal</tt> method must throw a
 <a href="../../../javax/xml/bind/MarshalException.html" title="class in javax.xml.bind"><code>MarshalException</code></a>. There exist two mechanisms
 to enable marshalling an instance that is not a JAXB element.
 One method is to wrap the instance as a value of a <a href="../../../javax/xml/bind/JAXBElement.html" title="class in javax.xml.bind"><code>JAXBElement</code></a>,
 and pass the wrapper element as the first parameter to
 a <tt>Marshaller.marshal</tt> method. For java to schema binding, it
 is also possible to simply annotate the instance's class with
 @<a href="../../../javax/xml/bind/annotation/XmlRootElement.html" title="annotation in javax.xml.bind.annotation"><code>XmlRootElement</code></a>.
 </blockquote>

 <p>
 <b>Encoding</b><br>
 </p><blockquote>
 By default, the Marshaller will use UTF-8 encoding when generating XML data
 to a <tt>java.io.OutputStream</tt>, or a <tt>java.io.Writer</tt>.  Use the
 <a href="../../../javax/xml/bind/Marshaller.html#setProperty(java.lang.String,%20java.lang.Object)"><code>setProperty</code></a> API to change the output
 encoding used during these marshal operations.  Client applications are
 expected to supply a valid character encoding name as defined in the
 <a href="http://www.w3.org/TR/2000/REC-xml-20001006#charencoding">W3C XML 1.0
 Recommendation</a> and supported by your
 <a href="http://java.sun.com/j2se/1.3/docs/api/java/lang/package-summary.html#charenc">
 Java Platform</a>.
 </blockquote>

 <p>
 <b>Validation and Well-Formedness</b><br>
 </p><blockquote>
 <p>
 Client applications are not required to validate the Java content tree prior
 to calling any of the marshal API's.  Furthermore, there is no requirement
 that the Java content tree be valid with respect to its original schema in
 order to marshal it back into XML data.  Different JAXB Providers will
 support marshalling invalid Java content trees at varying levels, however
 all JAXB Providers must be able to marshal a valid content tree back to
 XML data.  A JAXB Provider must throw a <tt>MarshalException</tt> when it
 is unable to complete the marshal operation due to invalid content.  Some
 JAXB Providers will fully allow marshalling invalid content, others will fail
 on the first validation error.
 </p><p>
 Even when schema validation is not explictly enabled for the marshal operation,
 it is possible that certain types of validation events will be detected
 during the operation.  Validation events will be reported to the registered
 event handler.  If the client application has not registered an event handler
 prior to invoking one of the marshal API's, then events will be delivered to
 a default event handler which will terminate the marshal operation after
 encountering the first error or fatal error. Note that for JAXB 2.0 and
 later versions, <a href="../../../javax/xml/bind/helpers/DefaultValidationEventHandler.html" title="class in javax.xml.bind.helpers"><code>DefaultValidationEventHandler</code></a> is
 no longer used.

 </p></blockquote>

 <p>
 <a name="supportedProps"></a>
 <b>Supported Properties</b><br>
 </p><blockquote>
 <p>
 All JAXB Providers are required to support the following set of properties.
 Some providers may support additional properties.
 </p><dl>
   <dt><tt>jaxb.encoding</tt> - value must be a java.lang.String</dt>
   <dd>The output encoding to use when marshalling the XML data.  The
               Marshaller will use "UTF-8" by default if this property is not
       specified.</dd>
   <dt><tt>jaxb.formatted.output</tt> - value must be a java.lang.Boolean</dt>
   <dd>This property controls whether or not the Marshaller will format
       the resulting XML data with line breaks and indentation.  A
       true value for this property indicates human readable indented
       xml data, while a false value indicates unformatted xml data.
       The Marshaller will default to false (unformatted) if this
       property is not specified.</dd>
   <dt><tt>jaxb.schemaLocation</tt> - value must be a java.lang.String</dt>
   <dd>This property allows the client application to specify an
       xsi:schemaLocation attribute in the generated XML data.  The format of
       the schemaLocation attribute value is discussed in an easy to
       understand, non-normative form in
       <a href="http://www.w3.org/TR/xmlschema-0/#schemaLocation">Section 5.6
       of the W3C XML Schema Part 0: Primer</a> and specified in
       <a href="http://www.w3.org/TR/xmlschema-1/#Instance_Document_Constructions">
       Section 2.6 of the W3C XML Schema Part 1: Structures</a>.</dd>
   <dt><tt>jaxb.noNamespaceSchemaLocation</tt> - value must be a java.lang.String</dt>
   <dd>This property allows the client application to specify an
       xsi:noNamespaceSchemaLocation attribute in the generated XML
       data.  The format of the schemaLocation attribute value is discussed in
       an easy to understand, non-normative form in
       <a href="http://www.w3.org/TR/xmlschema-0/#schemaLocation">Section 5.6
       of the W3C XML Schema Part 0: Primer</a> and specified in
       <a href="http://www.w3.org/TR/xmlschema-1/#Instance_Document_Constructions">
       Section 2.6 of the W3C XML Schema Part 1: Structures</a>.</dd>
   <dt><tt>jaxb.fragment</tt> - value must be a java.lang.Boolean</dt>
   <dd>This property determines whether or not document level events will be
       generated by the Marshaller.  If the property is not specified, the
       default is <tt>false</tt>. This property has different implications depending
       on which marshal api you are using - when this property is set to true:<br>
       <ul>
         <li><a href="../../../javax/xml/bind/Marshaller.html#marshal(java.lang.Object,%20org.xml.sax.ContentHandler)"><code>marshal(Object,ContentHandler)</code></a> - the Marshaller won't
             invoke <a href="../../../org/xml/sax/ContentHandler.html#startDocument()"><code>ContentHandler.startDocument()</code></a> and
             <a href="../../../org/xml/sax/ContentHandler.html#endDocument()"><code>ContentHandler.endDocument()</code></a>.</li>
         <li><a href="../../../javax/xml/bind/Marshaller.html#marshal(java.lang.Object,%20org.w3c.dom.Node)"><code>marshal(Object,Node)</code></a> - the property has no effect on this
             API.</li>
         <li><a href="../../../javax/xml/bind/Marshaller.html#marshal(java.lang.Object,%20java.io.OutputStream)"><code>marshal(Object,OutputStream)</code></a> - the Marshaller won't
             generate an xml declaration.</li>
         <li><a href="../../../javax/xml/bind/Marshaller.html#marshal(java.lang.Object,%20java.io.Writer)"><code>marshal(Object,Writer)</code></a> - the Marshaller won't
             generate an xml declaration.</li>
         <li><a href="../../../javax/xml/bind/Marshaller.html#marshal(java.lang.Object,%20javax.xml.transform.Result)"><code>marshal(Object,Result)</code></a> - depends on the kind of
             Result object, see semantics for Node, ContentHandler, and Stream APIs</li>
         <li><a href="../../../javax/xml/bind/Marshaller.html#marshal(java.lang.Object,%20javax.xml.stream.XMLEventWriter)"><code>marshal(Object,XMLEventWriter)</code></a> - the
             Marshaller will not generate <a href="../../../javax/xml/stream/XMLStreamConstants.html#START_DOCUMENT"><code>XMLStreamConstants.START_DOCUMENT</code></a> and
             <a href="../../../javax/xml/stream/XMLStreamConstants.html#END_DOCUMENT"><code>XMLStreamConstants.END_DOCUMENT</code></a> events.</li>
         <li><a href="../../../javax/xml/bind/Marshaller.html#marshal(java.lang.Object,%20javax.xml.stream.XMLStreamWriter)"><code>marshal(Object,XMLStreamWriter)</code></a> - the
             Marshaller will not generate <a href="../../../javax/xml/stream/XMLStreamConstants.html#START_DOCUMENT"><code>XMLStreamConstants.START_DOCUMENT</code></a> and
             <a href="../../../javax/xml/stream/XMLStreamConstants.html#END_DOCUMENT"><code>XMLStreamConstants.END_DOCUMENT</code></a> events.</li>
       </ul>
   </dd>
 </dl>
 </blockquote>

 <p>
 <a name="marshalEventCallback"></a>
 <b>Marshal Event Callbacks</b><br>
 </p><blockquote>
 "The <a href="../../../javax/xml/bind/Marshaller.html" title="interface in javax.xml.bind"><code>Marshaller</code></a> provides two styles of callback mechanisms
 that allow application specific processing during key points in the
 unmarshalling process.  In 'class defined' event callbacks, application
 specific code placed in JAXB mapped classes is triggered during
 marshalling.  'External listeners' allow for centralized processing
 of marshal events in one callback method rather than by type event callbacks.

 <p>
 Class defined event callback methods allow any JAXB mapped class to specify
 its own specific callback methods by defining methods with the following method signatures:
 </p><blockquote>
 <pre>   // Invoked by Marshaller after it has created an instance of this object.
   boolean beforeMarshal(Marshaller);

   // Invoked by Marshaller after it has marshalled all properties of this object.
   void afterMmarshal(Marshaller);
 </pre>
 </blockquote>
 The class defined event callback methods should be used when the callback method requires
 access to non-public methods and/or fields of the class.
 <p>
 The external listener callback mechanism enables the registration of a <a href="../../../javax/xml/bind/Marshaller.Listener.html" title="class in javax.xml.bind"><code>Marshaller.Listener</code></a>
 instance with a <a href="../../../javax/xml/bind/Marshaller.html#setListener(javax.xml.bind.Marshaller.Listener)"><code>setListener(Listener)</code></a>. The external listener receives all callback events,
 allowing for more centralized processing than per class defined callback methods.
 </p><p>
 The 'class defined' and external listener event callback methods are independent of each other,
 both can be called for one event. The invocation ordering when both listener callback methods exist is
 defined in <a href="../../../javax/xml/bind/Marshaller.Listener.html#beforeMarshal(java.lang.Object)"><code>Marshaller.Listener.beforeMarshal(Object)</code></a> and <a href="../../../javax/xml/bind/Marshaller.Listener.html#afterMarshal(java.lang.Object)"><code>Marshaller.Listener.afterMarshal(Object)</code></a>.
 </p><p>
 An event callback method throwing an exception terminates the current marshal process.
 </p></blockquote></div>
"""
        preprocessor = CodeTextPreprocessor()
        clean = preprocessor.clean_html_text(html_text=html_text)
        print(clean)



import re

from sekg.util.annotation import catch_exception

from sekg.constant.code import CodeEntityCategory, CodeEntityRelationCategory
from sekg.constant.constant import PropertyConstant
from sekg.graph.creator import NodeBuilder
from sekg.graph.exporter.graph_data import GraphData
from sekg.ir.doc.wrapper import MultiFieldDocumentCollection, MultiFieldDocument
from sekg.util.code import CodeElementNameUtil


class CodeElementGraphDataBuilder:
    VALID_IDENTITY_CHAR_SET = " ?,.QWERTYUIOPASDFGHJKLZXCVBNMqwertyuiopasdfghjklzxcvbnm1234567890_()<>[]"
    CHECK_CONSTANT_FILE_PATTERN = re.compile(r"[^A-Z_0-9]")

    def __init__(self, graph_data):
        self.graph_data = graph_data
        self.init_indexs()

    def init_indexs(self):
        self.graph_data.create_index_on_property("qualified_name", "short_description", "alias")

    def format_qualified_name(self, name):
        name = name.replace(", ", ",")
        for c in name:
            if c not in self.VALID_IDENTITY_CHAR_SET:
                return ""
        # todo: change this or add more judgement in here
        if "(" in name:
            start_name = name.split("(")
            if " " in start_name:
                return ""
        return name

    def parse_construct_to_javaparser_style(self, name):
        if "(" not in name:
            return name
        head = name.split("(")[0]
        tail = name.split("(")[1]

        return "{class_part}.<init>({parameter}".format(class_part=self.get_parent_name_for_api(head), parameter=tail)

    @staticmethod
    def get_simple_name_for_type(name):
        return name.split(".")[-1]

    @staticmethod
    def get_parent_name_for_api(name):
        split = name.split("(")[0].split(".")
        return ".".join(split[:-1])

    @staticmethod
    def get_parameter_num(method_name):

        if method_name.endswith("()"):
            return 0
        return method_name.count(",") + 1

    def find_match_method_by_actual_method_call(self, actual_method_call, method_nodes):
        if len(method_nodes) == 0:
            return None
        if len(method_nodes) == 1:
            return method_nodes[0]

        actual_parameter = self.get_parameter_num(actual_method_call)
        parameter_num_2_methods = {}

        for candidate_node in method_nodes:
            team = self.get_parameter_num(candidate_node["properties"]["qualified_name"])
            if team not in parameter_num_2_methods:
                parameter_num_2_methods[team] = []
            parameter_num_2_methods[team].append(candidate_node)

        if actual_parameter not in parameter_num_2_methods:
            print("match method fail for %r following candidate %r" % (
                actual_method_call,
                [candidate_node["properties"]["qualified_name"] for candidate_node in method_nodes]))

            return None

        if len(parameter_num_2_methods[actual_parameter]) > 1:
            print("match method fail for %r following candidate %r" % (
                actual_method_call,
                [candidate_node["properties"]["qualified_name"] for candidate_node in method_nodes]))

            return None

        return parameter_num_2_methods[actual_parameter][0]

    def add_base_value_entity_node(self, value_type, value_name, short_description="",
                                   entity_category=CodeEntityCategory.CATEGORY_VALUE, **extra_properties):
        qualified_name = "{type} {var_name}".format(type=value_type,
                                                    var_name=value_name)
        simple_name = "{type} {var_name}".format(type=self.get_simple_name_for_type(value_type),
                                                 var_name=value_name)
        if short_description == None:
            short_description = ""
        code_element = {
            "qualified_name": qualified_name,
            "simple_name": simple_name,
            "type": value_type,
            "value_name": value_name,
            "short_description": short_description,
        }
        for k, v in extra_properties.items():
            if k not in code_element:
                code_element[k] = v
        cat_labels = CodeEntityCategory.to_str_list(entity_category)

        builder = NodeBuilder().add_property(**code_element).add_entity_label().add_labels("code_element",
                                                                                           *cat_labels)

        new_field_node_id = self.graph_data.add_node_with_multi_primary_property(
            node_id=GraphData.UNASSIGNED_NODE_ID,
            node_labels=builder.get_labels(),
            node_properties=builder.get_properties(),
            primary_property_names=["qualified_name", "short_description"])

        type_node_id = self.add_type_node(value_type)

        self.graph_data.add_relation(new_field_node_id,
                                     CodeEntityRelationCategory.to_str(
                                         CodeEntityRelationCategory.RELATION_CATEGORY_TYPE_OF),
                                     type_node_id)

        return new_field_node_id

    def add_method_node(self, method_qualified_name, **extra_info):
        return self.add_normal_code_element_entity(qualified_name=method_qualified_name,
                                                   entity_category=CodeEntityCategory.CATEGORY_METHOD, **extra_info)

    def add_type_node(self, type_str):
        """
        add a new node stand for a type in GraphData.
        :param type_str:
        :return:
        """

        cat_label = CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_CLASS)
        builder = NodeBuilder().add_entity_label().add_labels("code_element", cat_label).add_one_property(
            "qualified_name",
            type_str)
        exist_json = self.graph_data.find_one_node_by_property(property_name="qualified_name",
                                                               property_value=type_str)

        if exist_json:
            return exist_json[self.graph_data.DEFAULT_KEY_NODE_ID]

        # print("add new type %s" % (type_str))

        type_node_id = self.graph_data.add_node(
            node_id=GraphData.UNASSIGNED_NODE_ID,
            node_labels=builder.get_labels(),
            node_properties=builder.get_properties(),
            primary_property_name="qualified_name")

        if type_node_id == GraphData.UNASSIGNED_NODE_ID:
            print("add new type fail: %r- %r" % (type_node_id, type_str))

        if "[]" in type_str:
            base_type_for_array = type_str.strip("[]")
            # print("array type- %s, base type- %s" % (type_str, base_type_for_array))
            base_type_node_id = self.add_type_node(base_type_for_array)

            self.graph_data.add_relation(type_node_id, "array of", base_type_node_id)

        return type_node_id

    def add_base_overrloading_method_node(self, simple_method_name):
        """
        add a new node stand for a base abstract method node. etc. java.lang.Math.abs, the method without parameters
        :param simple_method_name:  java.lang.Math.abs, the method without parameters
        :return:
        """

        cat_label = CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_BASE_OVERRIDE_METHOD)

        builder = NodeBuilder().add_entity_label().add_labels("code_element", cat_label).add_one_property(
            "qualified_name",
            simple_method_name)
        exist_json = self.graph_data.find_one_node_by_property(property_name="qualified_name",
                                                               property_value=simple_method_name)

        if exist_json:
            return exist_json[self.graph_data.DEFAULT_KEY_NODE_ID]

        new_node_id = self.graph_data.add_node(
            node_id=GraphData.UNASSIGNED_NODE_ID,
            node_labels=builder.get_labels(),
            node_properties=builder.get_properties(),
            primary_property_name="qualified_name")

        return new_node_id

    def add_normal_code_element_entity(self, qualified_name, entity_category, **extra_properties):
        """
        add the normal code element whose their qualified_name are unique
        :param qualified_name: the qualified_name of the code element
        :param entity_category:
        :param extra_properties:
        :return:
        """
        format_qualified_name = self.format_qualified_name(qualified_name)
        if not format_qualified_name:
            return GraphData.UNASSIGNED_NODE_ID

        code_element = {
            "qualified_name": format_qualified_name,
            "entity_category": entity_category
        }

        for k, v in extra_properties.items():
            if k not in code_element:
                code_element[k] = v

        cate_labels = CodeEntityCategory.to_str_list(entity_category)

        builder = NodeBuilder().add_entity_label().add_labels("code_element", *cate_labels).add_property(**code_element)

        node_id = self.graph_data.add_node(
            node_id=GraphData.UNASSIGNED_NODE_ID,
            node_labels=builder.get_labels(),
            node_properties=builder.get_properties(),
            primary_property_name="qualified_name")
        return node_id

    def add_relation_by_not_creating_entity(self, start_name, end_name, relation_type):
        start_name = self.format_qualified_name(start_name)
        end_name = self.format_qualified_name(end_name)

        if not start_name:
            print("start name not __valid, start name=%r" % start_name)
            return False
        if not end_name:
            print("end name not __valid, end name=%r" % end_name)
            return False

        start_node = self.graph_data.find_one_node_by_property(property_name="qualified_name",
                                                               property_value=start_name)
        if start_node is None:
            print("%r not found when creating relation" % start_name)
            return False

        start_node_id = start_node[GraphData.DEFAULT_KEY_NODE_ID]

        end_node = self.graph_data.find_one_node_by_property(property_name="qualified_name",
                                                             property_value=end_name)

        if end_node is None:
            print("%r not found when creating relation" % end_name)

            return False

        end_node_id = end_node[GraphData.DEFAULT_KEY_NODE_ID]

        self.graph_data.add_relation(start_node_id,
                                     CodeEntityRelationCategory.to_str(relation_type),
                                     end_node_id)

        return True

    def add_relation_by_creating_not_exist_entity(self, start_entity_name, end_entity_name, relation_type):
        if not start_entity_name or not end_entity_name:
            print("start name or end name is None")
            return False

        relation_type_str = CodeEntityRelationCategory.to_str(relation_type)

        start_name = self.format_qualified_name(start_entity_name)
        end_name = self.format_qualified_name(end_entity_name)

        if not start_name:
            print("start name not __valid, start name=%r" % start_entity_name)
            return False
        if not end_name:
            print("end name not __valid, end name=%r" % end_entity_name)
            return False

        start_node = self.graph_data.find_one_node_by_property(property_name="qualified_name",
                                                               property_value=start_name)

        if start_node is None:
            if "(" in start_name:
                print("add new method node for %r" % start_name)
                start_node_id = self.add_method_node(method_qualified_name=start_name)

            else:
                start_node_id = self.add_type_node(start_name)
        else:
            start_node_id = start_node[GraphData.DEFAULT_KEY_NODE_ID]

        end_node = self.graph_data.find_one_node_by_property(property_name="qualified_name",
                                                             property_value=end_name)

        if end_node is None:
            if "(" in end_name:
                print("add new method node for %r" % end_name)
                end_node_id = self.add_method_node(method_qualified_name=end_name)

            else:
                end_node_id = self.add_type_node(end_name)
        else:
            end_node_id = end_node[GraphData.DEFAULT_KEY_NODE_ID]

        self.graph_data.add_relation(start_node_id,
                                     relation_type_str,
                                     end_node_id)

        return True

    def add_method_use_class_relation(self, start_method_name, end_class_name):
        if not start_method_name or not end_class_name:
            print("start name or end name is None")
            return False
        raw_start_name = start_method_name

        relation_type_str = CodeEntityRelationCategory.to_str(
            CodeEntityRelationCategory.RELATION_CATEGORY_METHOD_IMPLEMENT_CODE_USE_CLASS)

        start_name = self.format_qualified_name(raw_start_name)
        start_node = self.graph_data.find_one_node_by_property(property_name="qualified_name",
                                                               property_value=start_name)

        if start_node is None:
            print("fail because of start node raw=%r format=%r" % (raw_start_name, start_name))
            return False
        type_node_id = self.add_type_node(end_class_name)

        self.graph_data.add_relation(start_node[GraphData.DEFAULT_KEY_NODE_ID],
                                     relation_type_str,
                                     type_node_id)

        return True

    def add_method_call_relation(self, start_name, end_name):
        if not start_name or not end_name:
            print("start name or end name is None")
            return False
        raw_start_name = start_name

        relation_type_str = CodeEntityRelationCategory.to_str(
            CodeEntityRelationCategory.RELATION_CATEGORY_METHOD_IMPLEMENT_CODE_CALL_METHOD)

        start_name = self.format_qualified_name(raw_start_name)
        start_node = self.graph_data.find_one_node_by_property(property_name="qualified_name",
                                                               property_value=start_name)

        if start_node is None:
            print("fail because of start node raw=%r format=%r" % (raw_start_name, start_name))
            return False

        end_node = self.graph_data.find_one_node_by_property(property_name="qualified_name",
                                                             property_value=end_name)

        if end_node is not None:
            self.graph_data.add_relation(start_node[GraphData.DEFAULT_KEY_NODE_ID],
                                         relation_type_str,
                                         end_node[GraphData.DEFAULT_KEY_NODE_ID])
            return True

        simple_name = end_name.split("(")[0] + "("
        parent_class_name = self.get_parent_name_for_api(simple_name)
        if not parent_class_name or parent_class_name == simple_name:
            return False

        method_ids = self.get_methods_belong_to_class_name(parent_class_name)

        candidate_nodes = []

        for method_id in method_ids:
            node_json = self.graph_data.get_node_info_dict(method_id)
            if simple_name not in node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]["qualified_name"]:
                continue
            candidate_nodes.append(node_json)

        end_node = self.find_match_method_by_actual_method_call(end_name, candidate_nodes)

        if end_node is None:
            end_abstract_method_node_id = self.add_base_overrloading_method_node(end_name.split("(")[0])
        else:
            end_abstract_method_node_id = end_node[GraphData.DEFAULT_KEY_NODE_ID]

        self.graph_data.add_relation(start_node[GraphData.DEFAULT_KEY_NODE_ID],
                                     relation_type_str,
                                     end_abstract_method_node_id)

        return True

    def build_abstract_overloading_relation(self):
        """
        for all methods, add a abstract new node stand for the overloading method with different parameters.
        its qualified name is simple method name without parameters. eg. java.lang.Math.abs
        :return:
        """
        print("start to build abstract_overloading_relation")
        print(self.graph_data)
        self.graph_data.print_graph_info()
        METHOD_CATE_STR = CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_METHOD)

        all_method_node_id_set = self.graph_data.get_node_ids_by_label(METHOD_CATE_STR)
        finish_node_id_set = set([])

        for node_id in all_method_node_id_set:
            if node_id in finish_node_id_set:
                continue

            node_json = self.graph_data.get_node_info_dict(node_id=node_id)

            method_qualified_name = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]["qualified_name"]
            if not self.format_qualified_name(method_qualified_name):
                finish_node_id_set.add(node_id)
                continue

            if "(" not in method_qualified_name:
                finish_node_id_set.add(node_id)
                continue

            base_overloading_method_simple_name = method_qualified_name.split("(")[0]
            new_base_method_node_id = self.add_base_overrloading_method_node(
                base_overloading_method_simple_name)

            finish_node_id_set.add(new_base_method_node_id)
            class_name = self.get_parent_name_for_api(base_overloading_method_simple_name)

            class_node = self.graph_data.find_one_node_by_property(
                property_name="qualified_name",
                property_value=class_name)

            if not class_node:
                continue

            candidate_method_ids_in_class = self.get_methods_belong_to_class(class_node[GraphData.DEFAULT_KEY_NODE_ID])

            for method_id in candidate_method_ids_in_class:
                candidate_method_json = self.graph_data.get_node_info_dict(method_id)
                candidate_qualified_name = candidate_method_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES][
                    "qualified_name"]

                if base_overloading_method_simple_name not in candidate_qualified_name:
                    continue
                finish_node_id_set.add(method_id)
                if method_id == new_base_method_node_id:
                    continue

                self.graph_data.add_relation(method_id,
                                             CodeEntityRelationCategory.to_str(
                                                 CodeEntityRelationCategory.RELATION_CATEGORY_METHOD_OVERLOADING),
                                             new_base_method_node_id)
        print(self.graph_data)
        self.graph_data.print_graph_info()
        print("end import abstract overloading_relation entity json")

    def build_belong_to_relation(self):
        """
        for all methods, add a abstract new node stand for the overloading method with different parameters.
        its qualified name is simple method name without parameters. eg. java.lang.Math.abs
        :return:
        """
        print("start build belong to relation")
        print(self.graph_data)
        self.graph_data.print_graph_info()

        VALID_LABELS = [

            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_METHOD),
            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_BASE_OVERRIDE_METHOD),
            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_FIELD_OF_CLASS),
            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_CLASS),
            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_INTERFACE),
            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_ENUM_CLASS),
            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_PACKAGE),
        ]

        add_new_parent_label = [CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_METHOD),
                                CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_BASE_OVERRIDE_METHOD),
                                CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_FIELD_OF_CLASS), ]
        for label in VALID_LABELS:
            all_node_id_set = self.graph_data.get_node_ids_by_label(label)

            for node_id in all_node_id_set:
                node_json = self.graph_data.get_node_info_dict(node_id=node_id)
                qualified_name = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]["qualified_name"]
                child_node_id = node_json[GraphData.DEFAULT_KEY_NODE_ID]

                parent_qualified_name = self.get_parent_name_for_api(qualified_name)

                if not parent_qualified_name or qualified_name == parent_qualified_name:
                    continue

                parent_node_json = self.graph_data.find_one_node_by_property("qualified_name", parent_qualified_name)

                if parent_node_json is None:
                    if label in add_new_parent_label:
                        parent_node_id = self.add_type_node(parent_qualified_name)
                        print("%r can't found, creating as class" % parent_qualified_name)
                    elif label == CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_PACKAGE):
                        parent_node_id = self.add_normal_code_element_entity(qualified_name=parent_qualified_name,
                                                                             entity_category=label)
                        print("%r can't found, creating as package" % parent_qualified_name)
                    elif parent_qualified_name == parent_qualified_name.lower():
                        if ".." in parent_qualified_name:
                            continue
                        parent_node_id = self.add_normal_code_element_entity(qualified_name=parent_qualified_name,
                                                                             entity_category=CodeEntityCategory.CATEGORY_PACKAGE)
                        print("%r can't found, creating as package" % parent_qualified_name)

                    else:
                        print("%r can't found" % parent_qualified_name)
                        continue
                else:
                    parent_node_id = parent_node_json[GraphData.DEFAULT_KEY_NODE_ID]

                self.graph_data.add_relation(child_node_id,
                                             CodeEntityRelationCategory.to_str(
                                                 CodeEntityRelationCategory.RELATION_CATEGORY_BELONG_TO),
                                             parent_node_id)

        print(self.graph_data)
        self.graph_data.print_graph_info()
        print("end build belong to relation")

    def build_value_subclass_relation(self):
        # todo: this is not useful, remove this part
        """
        for all methods, add a abstract new node stand for the overloading method with different parameters.
        its qualified name is simple method name without parameters. eg. java.lang.Math.abs
        :return:
        """
        print("start build value subclass relation")
        print(self.graph_data)
        self.graph_data.print_graph_info()

        all_node_id_set = self.graph_data.get_node_ids()
        finish_node_id_set = set([])

        VALUE_CAT_STR = CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_VALUE)

        for node_id in all_node_id_set:
            if node_id in finish_node_id_set:
                continue

            node_json = self.graph_data.get_node_info_dict(node_id=node_id)
            finish_node_id_set.add(node_id)

            if VALUE_CAT_STR not in node_json[GraphData.DEFAULT_KEY_NODE_LABELS]:
                continue

            parameter_qualified_name = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]["qualified_name"]

            short_description = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]["short_description"]
            value_type = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]["type"]
            value_name = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]["value_name"]

            if not short_description:
                base_value_node_id = self.add_base_value_entity_node(value_type=value_type,
                                                                     value_name=value_name,
                                                                     short_description="",
                                                                     entity_category=CodeEntityCategory.CATEGORY_VALUE)
            else:
                base_value_node_id = node_json[GraphData.DEFAULT_KEY_NODE_ID]

            finish_node_id_set.add(base_value_node_id)

            candidate_nodes = self.graph_data.find_nodes_by_property(
                property_name="qualified_name", property_value=parameter_qualified_name)

            for candidate_node in candidate_nodes:
                team_method_node_id = candidate_node[GraphData.DEFAULT_KEY_NODE_ID]
                if team_method_node_id in finish_node_id_set:
                    continue
                finish_node_id_set.add(team_method_node_id)
                if team_method_node_id == base_value_node_id:
                    continue

                self.graph_data.add_relation(team_method_node_id,
                                             CodeEntityRelationCategory.to_str(
                                                 CodeEntityRelationCategory.RELATION_CATEGORY_SUBCLASS_OF),
                                             base_value_node_id)

            super_base_value_node_id = self.add_base_value_entity_node(value_type=value_type,
                                                                       value_name="<V>",
                                                                       short_description="",
                                                                       entity_category=CodeEntityCategory.CATEGORY_VALUE)
            finish_node_id_set.add(super_base_value_node_id)

            self.graph_data.add_relation(base_value_node_id,
                                         CodeEntityRelationCategory.to_str(
                                             CodeEntityRelationCategory.RELATION_CATEGORY_SUBCLASS_OF),
                                         super_base_value_node_id)

        node_json = self.graph_data.find_one_node_by_properties(qualified_name="void <R>", short_description="")
        if node_json:
            self.graph_data.remove_node(node_json[GraphData.DEFAULT_KEY_NODE_ID])

        print(self.graph_data)
        self.graph_data.print_graph_info()
        print("end build value subclass relation")

    def get_methods_belong_to_class_name(self, class_qualified_name):
        class_json = self.graph_data.find_one_node_by_property(property_name="qualified_name",
                                                               property_value=class_qualified_name)
        if not class_json:
            return set([])

        class_id = class_json[GraphData.DEFAULT_KEY_NODE_ID]
        return self.get_methods_belong_to_class(class_id)

    def get_methods_belong_to_class(self, class_id):
        belong_to_relation_str = CodeEntityRelationCategory.to_str(
            CodeEntityRelationCategory.RELATION_CATEGORY_BELONG_TO)
        belong_to_relations = self.graph_data.get_relations(relation_type=belong_to_relation_str, end_id=class_id)
        valid_method_ids = set([])
        valid_method_labels = [
            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_METHOD),

            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_CONSTRUCT_METHOD),

            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_BASE_OVERRIDE_METHOD),

        ]

        for (method_node_id, _, _) in belong_to_relations:

            method_node_json = self.graph_data.get_node_info_dict(method_node_id)

            for label in valid_method_labels:
                if label in method_node_json[GraphData.DEFAULT_KEY_NODE_LABELS]:
                    valid_method_ids.add(method_node_id)
                    break

        return valid_method_ids

    def get_override_method_pairs(self, start_class_id, end_class_id):
        print("start try to locate override relation between %r - %r" % (start_class_id, end_class_id))

        start_class_node_json = self.graph_data.get_node_info_dict(start_class_id)
        end_class_node_json = self.graph_data.get_node_info_dict(end_class_id)

        if not start_class_node_json:
            return []
        if not end_class_node_json:
            return []

        start_class_name = start_class_node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES][
            GraphData.DEFAULT_KEY_PROPERTY_QUALIFIED_NAME]

        end_class_name = end_class_node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES][
            GraphData.DEFAULT_KEY_PROPERTY_QUALIFIED_NAME]

        print("try to locate override relation between %r - %r" % (start_class_name, end_class_name))

        method_ids_belong_to_start_class = self.get_methods_belong_to_class(start_class_id)

        override_method_pairs = []
        for method_node_id in method_ids_belong_to_start_class:

            method_node_json = self.graph_data.get_node_info_dict(method_node_id)
            if not method_node_json:
                continue

            if CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_CONSTRUCT_METHOD) in method_node_json[
                GraphData.DEFAULT_KEY_NODE_LABELS]:
                continue
            method_qualified_name = method_node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES][
                GraphData.DEFAULT_KEY_PROPERTY_QUALIFIED_NAME]
            if ".<init>" in method_qualified_name:
                continue
            if start_class_name not in method_qualified_name:
                print("error the method name %r don't contain class name %r" % (
                    method_qualified_name, start_class_name))
                continue

            new_method_qualified_name = method_qualified_name.replace(start_class_name, end_class_name)

            parent_method_node_json = self.graph_data.find_one_node_by_property(
                property_name=GraphData.DEFAULT_KEY_PROPERTY_QUALIFIED_NAME,
                property_value=new_method_qualified_name)

            if not parent_method_node_json:
                continue

            override_pair = (method_node_id, parent_method_node_json[GraphData.DEFAULT_KEY_NODE_ID])
            override_method_pairs.append(override_pair)
            print("found %r override %r" % (method_qualified_name, new_method_qualified_name))
        return override_method_pairs

    def build_override_relation(self):
        print("start build override relation")
        print(self.graph_data)
        self.graph_data.print_graph_info()

        VALID_LABELS = {
            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_CLASS),
            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_INTERFACE),
        }

        extends_relation_str = CodeEntityRelationCategory.to_str(
            CodeEntityRelationCategory.RELATION_CATEGORY_EXTENDS)
        implement_relation_str = CodeEntityRelationCategory.to_str(
            CodeEntityRelationCategory.RELATION_CATEGORY_IMPLEMENTS)
        extend_relation_list = self.graph_data.get_relations(relation_type=extends_relation_str)
        implement_relation_list = self.graph_data.get_relations(relation_type=implement_relation_str)

        all_relation_list = set(extend_relation_list).union(set(implement_relation_list))
        for (start_class_id, relation_type, end_class_id) in all_relation_list:
            print("try to find extends relation for %r %r %r" % (start_class_id, relation_type, end_class_id))
            override_method_pairs = self.get_override_method_pairs(start_class_id, end_class_id)

            for start_method_id, end_method_id in override_method_pairs:
                self.graph_data.add_relation(start_method_id,
                                             CodeEntityRelationCategory.to_str(
                                                 CodeEntityRelationCategory.RELATION_CATEGORY_METHOD_OVERRIDING),
                                             end_method_id)

        print(self.graph_data)
        self.graph_data.print_graph_info()
        print("end build override relation")

    def build_aliases_for_code_element(self):
        name_util = CodeElementNameUtil()
        all_node_id_set = self.graph_data.get_node_ids()

        for node_id in all_node_id_set:
            node_json = self.graph_data.get_node_info_dict(node_id=node_id)
            if "qualified_name" not in node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]:
                continue
            qualified_name = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]["qualified_name"]

            labels = node_json[GraphData.DEFAULT_KEY_NODE_LABELS]

            include_parent_name = False
            if (CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_METHOD) in labels and CodeEntityCategory.to_str(
                    CodeEntityCategory.CATEGORY_CONSTRUCT_METHOD) not in labels) or CodeEntityCategory.to_str(
                CodeEntityCategory.CATEGORY_ENUM_CONSTANTS) in labels or CodeEntityCategory.to_str(
                CodeEntityCategory.CATEGORY_FIELD_OF_CLASS) in labels or CodeEntityCategory.to_str(
                CodeEntityCategory.CATEGORY_BASE_OVERRIDE_METHOD) in labels:
                include_parent_name = True

            name_list = name_util.generate_aliases(qualified_name=qualified_name,
                                                   include_simple_parent_name=include_parent_name)

            if "simple_name" in node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]:
                name_list.append(node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]["simple_name"])

            node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]["alias"] = list(set(name_list))

        self.graph_data.refresh_indexer()

    def add_source_label(self, label):
        print("adding the source label %r to nodes in graph" % label)
        self.graph_data.add_label_to_all(label)

    def build(self):
        """
        get the graph after builder
        :return: the graph data build successfully
        """
        return self.graph_data

    @catch_exception
    def build_use_jdk_constant_field_relation_from_code_doc(self, document_collection: MultiFieldDocumentCollection):
        if document_collection == None:
            raise Exception("The code document collection can't be None")

        constant_nodes = []
        for node_id in self.graph_data.get_node_ids_by_label(
                CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_FIELD_OF_CLASS)):
            node_json = self.graph_data.get_node_info_dict(node_id=node_id)
            if not node_json:
                continue
            node_properties = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]
            simple_constant_name = node_properties["qualified_name"].split(".")[-1]

            if len(node_properties["qualified_name"].split(".")) <= 2:
                continue
            parent_class_name = node_properties["qualified_name"].split(".")[-2]
            constant_nodes.append((node_id, parent_class_name + "." + simple_constant_name))
        # for t in constant_nodes:
        #     print(t)

        for node_id in self.graph_data.get_node_ids_by_label(
                CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_METHOD)):
            doc: MultiFieldDocument = document_collection.get_by_id(node_id)
            if doc == None:
                continue
            code_text = doc.get_doc_text_by_field("code")

            for constant_node_id, constant_node_name in constant_nodes:

                if constant_node_name not in code_text:
                    continue

                self.graph_data.add_relation(startId=node_id, endId=constant_node_id,
                                             relationType=CodeEntityRelationCategory.to_str(
                                                 CodeEntityRelationCategory.RELATION_CATEGORY_METHOD_IMPLEMENT_CODE_CALL_FIELD
                                             ))

        return self.graph_data


    def export_code_document_collection(self, code_doc_collection_path=None):
        """
        export a document collection containing all code
        :param code_doc_collection_path:
        :return:
        """
        collection = MultiFieldDocumentCollection()

        for node_id in self.graph_data.get_node_ids():
            node_json = self.graph_data.get_node_info_dict(node_id=node_id)

            properties = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]
            ## the code is empty is acceptable

            code = properties.get("code", None)
            if code != None:
                name = properties.get("qualified_name", "")
                if name == "":
                    continue

                doc = MultiFieldDocument(id=node_id, name=name)
                doc.add_field("code", code)
                doc.add_field("qualified_name", name)

                collection.add_document(document=doc)
        if code_doc_collection_path!=None:
            collection.save(code_doc_collection_path)
        return collection



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import asyncio
import pickle
import time
import traceback
from pathlib import Path

import aiohttp
import async_timeout

from sekg.util.annotation import catch_exception
from sekg.wiki.WikiDataItem import WikiDataItem


# if sys.platform == 'win32':
#     loop = asyncio.ProactorEventLoop()
#     asyncio.set_event_loop(loop)

# import wikipedia


class AsyncWikiSearcher:
    API_URL = 'https://www.wikidata.org/w/api.php'

    def __init__(self, proxy_server=None, pool_size=63, stride=60):
        self.proxy_server = proxy_server
        self.semaphore = asyncio.Semaphore(pool_size)
        self.title_cache = {}
        self.item_cache = {}
        self.stride = stride

    async def __fetch_titles(self, query, limit=5):
        if query in self.title_cache:
            return self.title_cache[query]
        params = {
            'action': 'query',
            'format': 'json',
            'list': 'search',
            'srprop': 'snippet',
            'srlimit': limit,
            'srsearch': query
        }
        try:
            async with self.semaphore:
                async with aiohttp.ClientSession() as session:
                    with async_timeout.timeout(10):
                        async with session.get(self.API_URL, params=params, proxy=self.proxy_server) as response:
                            json_data = await response.json()
                            result = [
                                {"pageid": item["pageid"], "title": item["title"], "snippet": item.get("snippet", "")}
                                for item in json_data["query"]["search"]]
                            self.title_cache[query] = result
                            # print("[Done] query: {}".format(query))
                            return result
        except Exception:
            print("[Failed] query: {}".format(query))
            traceback.print_exc()
            return []

    async def __fetch_entity(self, title):
        if title in self.item_cache:
            return self.item_cache[title]
        params = {
            'ids': title,
            'format': 'json',
            'action': 'wbgetentities'
        }
        try:
            async with self.semaphore:
                async with aiohttp.ClientSession() as session:
                    with async_timeout.timeout(10):
                        async with session.get(self.API_URL, params=params, proxy=self.proxy_server) as response:
                            json_data = await response.json()
                            self.item_cache[title] = WikiDataItem(title,
                                                                  init_at_once=False).init_wikidata_item_from_json(
                                json_data)
                            # print("[Done] title: {}".format(title))
                            return json_data
        except Exception:
            print("[Failed] title: {}".format(title))
            traceback.print_exc()
            return {}

    def save(self, title_save_path=None, item_save_path=None):
        print("Save...")
        if title_save_path is not None:
            with Path(title_save_path).open("wb") as f:
                pickle.dump(self.title_cache, f)
        if item_save_path is not None:
            with Path(item_save_path).open("wb") as f:
                pickle.dump(self.item_cache, f)

    def clear(self):
        del self.item_cache
        del self.title_cache
        self.item_cache = {}
        self.title_cache = {}

    def search_title(self, queries):
        loop = asyncio.get_event_loop()
        tasks = [self.__fetch_titles(q) for q in queries]
        loop.run_until_complete(asyncio.gather(*tasks))
        return self.title_cache

    def fetch_item(self, ids):
        loop = asyncio.get_event_loop()
        tasks = [self.__fetch_entity(_id) for _id in ids]
        loop.run_until_complete(asyncio.gather(*tasks))
        # for k, v in self.item_cache:
        #     self.item_cache[k] = WikiDataItem(k, init_at_once=False).init_wikidata_item_from_json(v)
        return self.item_cache

    @catch_exception
    def init_from_cache(self, title_save_path=None, item_save_path=None):
        print("Init from cache...")
        if title_save_path is not None and Path(title_save_path).exists():
            with Path(title_save_path).open("rb") as f:
                title_cache = pickle.load(f)
                self.title_cache = dict(self.title_cache, **title_cache)
        if item_save_path is not None and Path(item_save_path).exists():
            with Path(item_save_path).open("rb") as f:
                item_cache = pickle.load(f)
                self.item_cache = dict(self.item_cache, **item_cache)

    def item_cache_size(self):
        return len(self.item_cache)

    def title_cache_size(self):
        return len(self.title_cache)

    def get_item_cache(self):
        return self.item_cache

    def get_title_cache(self):
        return self.title_cache

    def __repr__(self):
        return "<AsyncWikiSearcher title=%r item=%r>" % (self.title_cache_size(), self.item_cache_size())


if __name__ == '__main__':
    searcher = AsyncWikiSearcher(proxy_server="http://127.0.0.1:1080")
    start = time.time()
    with Path("terms.txt").open("r", encoding="utf-8") as f:
        terms = [line.strip() for line in f]
    searcher.search_title(["apple", "orange"])
    searcher.save(title_save_path="titles.bin")
    ids = {item["title"] for v in searcher.title_cache.values() for item in v}
    searcher.fetch_item(ids)
    searcher.save(item_save_path="items.bin")
    # result = searcher.summary()
    end = time.time()
    print("cost time:", end - start)



import re


class CodeElementNameUtil:
    PATTERN_2_4 = re.compile(r'([A-Za-z])([24])([A-CE-Za-ce-z])')
    PATTERN_split = re.compile(r'([A-Z]+)([A-Z][a-z0-9]+)')
    PATTERN_split_num = re.compile(r'([0-9]?[A-Z]+)')

    def get_simple_name_with_parent(self, name):
        if not name:
            return None
        team_name = name.split("(")[0]
        split_names = team_name.split(".")
        if len(split_names) <= 1:
            return split_names[-1]

        child = split_names[-1].strip()
        parent = split_names[-2].strip()

        return parent + "." + child

    def simplify(self, name):
        """
        get the simple name for class, method, field, eg. java.util.ArrayList->ArrayList
        :param name:
        :return:
        """
        if not name:
            return None
        team_name = name.split("(")[0]
        simple_name = team_name.split(".")[-1].strip()

        return simple_name

    def uncamelize_from_simple_name(self, name):
        """
        uncamel from simple name of one name, rg. java.util.ArrayList->Array List
        :param name:
        :return:
        """
        if not name:
            return None
        simple_name = self.simplify(name)

        return self.uncamelize(simple_name)

    def uncamelize(self, name):
        """
        uncamel one name
        :param name: the camel styple name(include underline)
        :return:
        """
        if not name:
            return None
        # sub = re.sub(r'([A-Za-z])([24])([A-CE-Za-ce-z])', r'\1 \2 \3', name).strip()
        sub = re.sub(self.PATTERN_2_4, r'\1 \2 \3', name).strip()
        sub = re.sub(r'_', " ", sub)
        # sub = re.sub(r'([A-Z]+)([A-Z][a-z0-9]+)', r'\1 \2', sub)
        sub = re.sub(self.PATTERN_split, r'\1 \2', sub)
        # sub = re.sub(r'([0-9]?[A-Z]+)', r' \1', sub)
        sub = re.sub(self.PATTERN_split_num, r' \1', sub)
        sub = re.sub(r'\s+', " ", sub).strip()
        return sub

    def uncamelize_by_stemming(self, name):
        """
        uncamelzie the name and remove last num, eg. Student1->Student, JavaParser3->Java Parser
        :param name:
        :return:
        """
        # todo: improve this method to fix more situation, has some error for Path1->Path
        name = re.sub(r'([0-9]+)$', '', name)
        name = self.uncamelize(name)
        if not name:
            return None
        sub = self.match_numer_first_and_middle(name)
        if sub:
            return sub
        # self.merge_after_uncamelize_and_stemm(name)
        return name

    def match_numer_first_and_middle(self, name):
        number_first = re.compile(r'(^[0-9]+[A-Z]+)', re.IGNORECASE).findall(name)
        if number_first:
            return number_first[0]
        else:
            number_middle = re.compile(r'([A-Z]+[24][A-Z]+)', re.IGNORECASE).findall(name.replace(" ", ""))
            if number_middle:
                return number_middle[0]
            else:
                return None

    def generate_aliases(self, qualified_name, include_simple_parent_name=False):
        if not qualified_name:
            return []

        simple_name = self.simplify(qualified_name)
        separate_name = self.uncamelize_from_simple_name(simple_name)
        name_list = [simple_name, separate_name]

        if include_simple_parent_name:
            name_list.append(self.get_simple_name_with_parent(qualified_name))

        name_list = [name for name in name_list if name]

        return list(set(name_list))


class ConceptElementNameUtil:
    PATTERN_2_4 = re.compile(r'([A-Za-z])([24])([A-CE-Za-ce-z])')
    PATTERN_split = re.compile(r'([A-Z]+)([A-Z][a-z0-9]+)')
    PATTERN_split_num = re.compile(r'([0-9]?[A-Z]+)')

    def get_simple_name_with_parent(self, name):
        if not name:
            return None
        team_name = name.split("(")[0]
        split_names = team_name.split(".")
        if len(split_names) <= 1:
            return split_names[-1]

        child = split_names[-1].strip()
        parent = split_names[-2].strip()

        return parent + "." + child

    def simplify(self, name):
        """
        get the simple name for class, method, field, eg. java.util.ArrayList->ArrayList
        :param name:
        :return:
        """
        if not name:
            return None
        team_name = name.split("(")[0]
        simple_name = team_name.split(".")[-1].strip()

        return simple_name

    def uncamelize_from_simple_name(self, name):
        """
        uncamel from simple name of one name, rg. java.util.ArrayList->Array List
        :param name:
        :return:
        """
        if not name:
            return None
        simple_name = self.simplify(name)

        return self.uncamelize(simple_name)

    def uncamelize(self, name):
        """
        uncamel one name
        :param name: the camel styple name(include underline)
        :return:
        """
        if not name:
            return None
        # sub = re.sub(r'([A-Za-z])([24])([A-CE-Za-ce-z])', r'\1 \2 \3', name).strip()
        sub = re.sub(self.PATTERN_2_4, r'\1 \2 \3', name).strip()
        sub = re.sub(r'_', " ", sub)
        # sub = re.sub(r'([A-Z]+)([A-Z][a-z0-9]+)', r'\1 \2', sub)
        sub = re.sub(self.PATTERN_split, r'\1 \2', sub)
        # sub = re.sub(r'([0-9]?[A-Z]+)', r' \1', sub)
        sub = re.sub(self.PATTERN_split_num, r' \1', sub)
        sub = re.sub(r'\s+', " ", sub).strip()
        return sub

    def uncamelize_by_stemming(self, name):
        """
        uncamelzie the name and remove last num, eg. Student1->Student, JavaParser3->Java Parser
        :param name:
        :return:
        """
        # todo: improve this method to fix more situation, has some error for Path1->Path
        name = re.sub(r'([0-9]+)$', '', name)
        name = self.uncamelize(name)
        if not name:
            return None
        sub = self.match_numer_first_and_middle(name)
        if sub:
            return sub
        # self.merge_after_uncamelize_and_stemm(name)
        return name

    def match_numer_first_and_middle(self, name):
        number_first = re.compile(r'(^[0-9]+[A-Z]+)', re.IGNORECASE).findall(name)
        if number_first:
            return number_first[0]
        else:
            number_middle = re.compile(r'([A-Z]+[24][A-Z]+)', re.IGNORECASE).findall(name.replace(" ", ""))
            if number_middle:
                return number_middle[0]
            else:
                return None

    def generate_aliases(self, qualified_name, vocabulary=None, abbreviation=False):
        if not qualified_name:
            return []

        simple_name = self.simplify(qualified_name)
        of_names = self.deal_with_adj(simple_name)
        separate_name = self.uncamelize_from_simple_name(simple_name)
        name_deal_number = self.uncamelize_by_stemming(simple_name)
        combined_name = simple_name.lower().replace("-", "").replace("\\", "").replace(" ", "")

        name_list = [simple_name, separate_name, name_deal_number, combined_name]
        name_list.extend(of_names)

        if abbreviation:
            result = self.generate_all_abbreviation_names(separate_name, vocabulary)
            name_list.extend(result)

        name_list = [name for name in name_list if name]

        return list(set(name_list))

    def generate_all_abbreviation_names(self, separate_name, vocabulary):
        result = []
        part_abbreviation = []
        full_abbreviation_name = self.get_abbreviation(separate_name)
        abbreviation_name = self.get_abbreviation(separate_name, full_link=False)
        if vocabulary != None:
            part_abbreviation = self.get_part_abbreviation(separate_name, vocabulary)
        result.append(full_abbreviation_name)
        result.append(abbreviation_name)
        result.extend(part_abbreviation)

        return result

    def deal_with_adj(self, name):
        result = []
        seperate_words = [" of ", " Of "]

        # A of B type
        for s_w in seperate_words:
            if s_w in name:
                words = name.split(s_w)
                if len(words) != 2:
                    continue
                child = words[0]
                parent = words[1]
                result.append((parent + " " + child).replace("  ", " "))

        seperate_words = ["'s ", "' "]

        # A's B => A B
        for s_w in seperate_words:
            if s_w in name:
                words = name.split(s_w)
                if len(words) != 2:
                    continue
                parent = words[0]
                child = words[1]

                result.append((parent + " " + child).replace("  ", " "))
                result.append((child + " of " + parent).replace("  ", " "))

        return result

    def get_abbreviation(self, separate_name, full_link=True):
        separate_name_list = separate_name.lower().split(" ")
        if len(separate_name_list) <= 1:
            return separate_name
        if full_link:
            abbreviation_list = [name[0].upper() for name in separate_name_list]
            return "".join(abbreviation_list)
        else:
            abbreviation_list = [name[0].upper() for name in separate_name_list if
                                 name not in ["of", "the", "this", "a", "that"]]
            return "".join(abbreviation_list)

    def get_part_abbreviation(self, separate_name, vocabulary):
        """
        :param separate_name: term name
        :param vocabulary: a list of prase
        :return: a list of abbreviation
        """
        separate_name_list = separate_name.split(" ")
        return_list = []
        if len(separate_name_list) > 1:
            for index in range(len(separate_name_list) - 1):
                for inner_index in range(index + 1, len(separate_name_list)):
                    prase = " ".join(separate_name_list[index:inner_index + 1])
                    if prase in vocabulary:
                        abbreviation_list = [name[0].upper() for name in separate_name_list[index:inner_index + 1]]
                        abbreviation = separate_name.replace(prase, "".join(abbreviation_list))
                        return_list.append(abbreviation)
            return return_list
        else:
            return [separate_name]



from unittest import TestCase

from sekg.text.extractor.domain_entity.identifier_util import IdentifierInfoExtractor


class TestIdentifierHelper(TestCase):
    def test_extract_from_variable(self):
        extractor = IdentifierInfoExtractor()

        test_cases = [
            ("String filePath", {"file", "path"}),
            ("String srcFile", {"src File"}),
            ("String srcFile1", {"src File"}),
            ("String i1", {}),
            ("String 3d", {}),
            ("String AA1", {}),
            ("String HTML5", {}),
            ("String HTML", {}),
            # 反驼峰式， 对于 全大写加数字，全大写和数字需要拼接在一起。如HTML5
        ]
        for name, answer in test_cases:
            print("extract for method name", name)
            print(extractor.extract_from_variable(name))
            # self.assertEqual(answer, extractor.extract_from_variable(name))

    def test_extract_from_class_name(self):
        extractor = IdentifierInfoExtractor()
        # todo: fix the case problem
        test_cases = [
            ("String", {"string"}),
            ("StringSrcFile", {"string Src File"}),
            ("String3D", {"string 3d"}),
            ("UserID", {"user ID"}),
        ]
        for name, answer in test_cases:
            print("extract for method name", name)
            print(extractor.extract_from_class_name(name))
            # self.assertEqual(answer, extractor.extract_from_class_name(name))

    def test_extract_from_method_for_operate_relation(self):
        self.fail()

    def test_pos_tag_for_method(self):
        extractor = IdentifierInfoExtractor()
        # todo: fix the case problem
        test_cases = [
            ("StringBuilder.build(int)", [('build', 'VERB')]),
            ("StringBuilder.toString()", [('to', 'PART'), ('string', 'VERB')]),
            ("StringBuilder.getConn()", [('to', 'PART'), ('string', 'VERB')]),
            ("List.getUserId", ({"Container"}, set(["delete"]))),
            ("List.id", ({"Container"}, set(["delete"]))),

            ("List.findOrCreateCon", ({"Container"}, set(["delete"]))),
            ("List.buttonContainer", ({"Container"}, set(["delete"]))),

            ("List.deleteFromContainer", ({"Container"}, set(["delete"]))),
            ("List.deleteAndRemoveFromContainer", ({"Container"}, set(["delete"]))),

            ("String3D.convertStrToInt(sting,int)", ({"string"}, set())),
            ("String3D.convertStrToLargeImg(sting,int)", ({"string"}, set())),

            ("Integer.parseToInt", ({"int"}, set(["parse"]))),
            ("Integer.parseToLongInt", ({"int"}, set(["parse"]))),
            ("Integer.parseToLong", ({"int"}, set(["parse"]))),
            ("Integer.parseLong", ({"int"}, set(["parse"]))),
            ("Integer.parseInt", ({"int"}, set(["parse"]))),
            ("Integer.parseInteger", ({"int"}, set(["parse"]))),

            ("StringBuilder.id2score()", [('to', 'PART'), ('string', 'VERB')]),
            ("StringBuilder.name2score()", [('to', 'PART'), ('string', 'VERB')]),
            ("List.deleteIntFromContainer", ({"Container"}, set(["delete"]))),
            ("List.deleteIntFromContainerByID", ({"Container"}, set(["delete"]))),
            ("List.isExist", ({"Container"}, set(["delete"]))),
            ("List.isRed", ({"Container"}, set(["delete"]))),

            ("List.html", ({"Container"}, set(["delete"]))),
            ("List.list", ({"Container"}, set(["delete"]))),
            ("List.listName", ({"Container"}, set(["delete"]))),
            ("List.setButtonEnable", ({"Container"}, set(["delete"]))),
            ("List.setButtonEnabled", ({"Container"}, set(["delete"]))),
            ("List.makeButtonEnabled", ({"Container"}, set(["delete"]))),
            ("List.makeButtonEnable", ({"Container"}, set(["delete"]))),

            ("List.makeButtonUnclickable", ({"Container"}, set(["delete"]))),
            ("List.quickSort", ({"Container"}, set(["delete"]))),
            ("List.quicksort", ({"Container"}, set(["delete"]))),
            ("List.mergeSort", ({"Container"}, set(["delete"]))),
            ("List.mergesort", ({"Container"}, set(["delete"]))),
            ("List.BM25", ({"Container"}, set(["delete"]))),
            ("List.createContainer", ({"Container"}, set(["delete"]))),
            ("List.len", ({"Container"}, set(["delete"]))),
            ("List.size", ({"Container"}, set(["delete"]))),

            ("List.invokeOtherMethods", ({"Container"}, set(["delete"]))),

        ]
        for name, answer in test_cases:
            print("extract for method name", name)
            extractor.pos_tag_for_method_name(name)

            # self.assertEqual(answer, extractor.pos_tag_for_method_name(name))

    def test_extract_for_method(self):
        extractor = IdentifierInfoExtractor()
        # todo: fix the case problem
        test_cases = [
            ("StringBuilder.build(int)", [('build', 'VERB')]),
            ("StringBuilder.toString()", [('to', 'PART'), ('string', 'VERB')]),
            ("StringBuilder.getConn()", [('to', 'PART'), ('string', 'VERB')]),
            ("List.getUserId", ({"Container"}, set(["delete"]))),
            ("List.id", ({"Container"}, set(["delete"]))),

            ("List.findOrCreateCon", ({"Container"}, set(["delete"]))),
            ("List.buttonContainer", ({"Container"}, set(["delete"]))),

            ("List.deleteFromContainer", ({"Container"}, set(["delete"]))),
            ("List.deleteAndRemoveFromContainer", ({"Container"}, set(["delete"]))),

            ("String3D.convertStrToInt(sting,int)", ({"string"}, set())),
            ("String3D.convertStrToLargeImg(sting,int)", ({"string"}, set())),

            ("Integer.parseToInt", ({"int"}, set(["parse"]))),
            ("Integer.parseToLongInt", ({"int"}, set(["parse"]))),
            ("Integer.parseToLong", ({"int"}, set(["parse"]))),
            ("Integer.parseLong", ({"int"}, set(["parse"]))),
            ("Integer.parseInt", ({"int"}, set(["parse"]))),
            ("Integer.parseInteger", ({"int"}, set(["parse"]))),

            ("StringBuilder.id2score()", [('to', 'PART'), ('string', 'VERB')]),
            ("StringBuilder.name2score()", [('to', 'PART'), ('string', 'VERB')]),
            ("List.deleteIntFromContainer", ({"Container"}, set(["delete"]))),
            ("List.deleteIntFromContainerByID", ({"Container"}, set(["delete"]))),
            ("List.isExist", ({"Container"}, set(["delete"]))),
            ("List.isRed", ({"Container"}, set(["delete"]))),

            ("List.html", ({"Container"}, set(["delete"]))),
            ("List.list", ({"Container"}, set(["delete"]))),
            ("List.listName", ({"Container"}, set(["delete"]))),
            ("List.setButtonEnable", ({"Container"}, set(["delete"]))),
            ("List.setButtonEnabled", ({"Container"}, set(["delete"]))),
            ("List.makeButtonEnabled", ({"Container"}, set(["delete"]))),
            ("List.makeButtonEnable", ({"Container"}, set(["delete"]))),

            ("List.makeButtonUnclickable", ({"Container"}, set(["delete"]))),
            ("List.quickSort", ({"Container"}, set(["delete"]))),
            ("List.quicksort", ({"Container"}, set(["delete"]))),
            ("List.mergeSort", ({"Container"}, set(["delete"]))),
            ("List.mergesort", ({"Container"}, set(["delete"]))),
            ("List.BM25", ({"Container"}, set(["delete"]))),
            ("List.createContainer", ({"Container"}, set(["delete"]))),
            ("List.len", ({"Container"}, set(["delete"]))),
            ("List.size", ({"Container"}, set(["delete"]))),

            ("List.invokeOtherMethods", ({"Container"}, set(["delete"]))),

        ]
        for name, answer in test_cases:
            print("extract for method name", name)
            print(extractor.extract_from_method_name(name))

            # self.assertEqual(answer, extractor.pos_tag_for_method_name(name))



from unittest import TestCase

from sekg.text.extractor.domain_entity.entity_extractor import EntityExtractor


class TestEntityExtractor(TestCase):
    def test_extract_from_sentence(self):
        sent = """Specifically, workspaces are designed for cyclical workloads - such as training neural networks - as
            they allow for off-heap memory reuse (instead of continually allocating and deallocating memory on each iteration
            of the loop). """

        extractor = EntityExtractor()
        terms = extractor.extract_from_sentence(sent)
        for term in terms:
            print(term)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import itertools
import re
from enum import Enum, unique

# todo: need refactor, some of the relation name are only for the code, eg. mention in
from sekg.text.extractor.domain_entity.nlp_util import SpacyNLPFactory


@unique
class RelationType(Enum):
    IS_A = "is a"
    REPRESENT = "represent"
    SUBCLASS_OF = "subclass of"
    DERIVED_FROM = "part of"
    MENTION_IN_COMENT = "mention in comment"
    MENTION_IN_INSIDE_COMENT = "mention in inside comment"
    MENTION_IN_STRING_LITERAL = "mention in string literal"
    MENTION_IN_SHORT_DESCRIPTION = "mention in short description"
    MENTION = "mention"

    HAS_OPERATION = "has operation"
    INSTANCE_OF = "instance of"
    OPERATE = "operate"
    CAN_BE_OPERATED = "can be operated"

    NAME_MENTION = "name mention"


class RelationDetector:
    def __init__(self):
        self.cache = {}
        self.nlp = SpacyNLPFactory.create_spacy_nlp_for_domain_extractor()
        self.pos_cache = {}

    def expand(self, terms):
        terms = set(terms)
        expanded = set()
        addition = set()
        for term in terms:
            tmp = re.sub(r'( and | or )', r' $ ', term)
            parts = tmp.split(" $ ")
            if len(parts) > 1:
                i = 0
                size = len(parts)
                while i < size - 1:
                    p = parts[i].split()
                    n = parts[i + 1].split()
                    if len(p) < len(n):
                        tmp = " ".join(p + n[len(p):])
                        addition.add(tmp)
                        if tmp not in self.cache:
                            self.cache[tmp] = set()
                        self.cache[tmp].add(term)
                        addition.add(parts[i + 1])
                        if parts[i + 1] not in self.cache:
                            self.cache[parts[i + 1]] = set()
                        self.cache[parts[i + 1]].add(term)
                        i += 2
                    else:
                        addition.add(parts[i])
                        if parts[i] not in self.cache:
                            self.cache[parts[i]] = set()
                        self.cache[parts[i]].add(term)
                        i += 1
                        if i == size - 1:
                            addition.add(parts[i])
                            if parts[i] not in self.cache:
                                self.cache[parts[i]] = set()
                            self.cache[parts[i]].add(term)
                expanded.add(term)

            tmp = re.sub(r'( of | \'s )', r' $ ', term)
            parts = tmp.split(" $ ")
            if len(parts) == 2:
                addition.add(parts[0])
                if parts[0] not in self.cache:
                    self.cache[parts[0]] = set()
                self.cache[parts[0]].add(term)
                addition.add(parts[1])
                if parts[1] not in self.cache:
                    self.cache[parts[1]] = set()
                self.cache[parts[1]].add(term)
                expanded.add(term)

        terms.update(addition)
        return expanded

    def detect_relation_by_starfix(self, terms):
        """[summary]

        detect relation by *fix (prefix, infix, suffix)

        Arguments:
            terms {[type]} -- [description]

        Returns:
            [type] -- [description]
        """
        expanded = self.expand(terms)
        relations = set()
        for term1, term2 in itertools.combinations(terms - expanded, 2):
            short_term, long_term = (term1.lower(), term2.lower()) if len(term1) <= len(term2) else (
                term2.lower(), term1.lower())

            if long_term.endswith(" " + short_term):
                if long_term in self.pos_cache:
                    long_term_pos = self.pos_cache[long_term]
                else:
                    long_term_pos = tuple([token.pos_ for token in self.nlp(long_term)])
                    # TODO: FIX THIS, ONLY ADD MOST NP
                    self.pos_cache[long_term] = long_term_pos
                if long_term_pos[-1] in {"NOUN", "PROPN"}:
                    relations.add((long_term, RelationType.IS_A.value, short_term))
            elif " {} ".format(short_term) in long_term or long_term.startswith(short_term + " "):
                if long_term in self.pos_cache:
                    long_term_pos = self.pos_cache[long_term]
                else:
                    long_term_pos = tuple([token.pos_ for token in self.nlp(long_term)])
                    self.pos_cache[long_term] = long_term_pos
                if long_term_pos[long_term.split().index(short_term.split()[-1])] in {"NOUN", "PROPN"}:
                    relations.add((long_term, RelationType.DERIVED_FROM.value, short_term))

        for k, v in self.cache.items():
            for term in v:
                relations.add((term, RelationType.DERIVED_FROM.value, k))

        relations = set(filter(lambda x: x[0] != x[2], relations))
        return relations


if __name__ == "__main__":
    print(RelationType.IS_A.value)



from unittest import TestCase

from sekg.text.extractor.domain_entity.word_util import WordUtil


class TestWordUtil(TestCase):
    def test_couldBeVerb(self):
        self.fail()

    def test_couldBeNoun(self):
        self.fail()

    def test_couldBeADJ(self):
        test_cases = [
            ("red", True),
            ("file", False),
            ("readable", True),

        ]
        for word, result in test_cases:
            self.assertEqual(result, WordUtil.couldBeADJ(word))



import spacy
import neuralcoref
import operator
import re

from sekg.util.code import CodeElementNameUtil


class ReferenceResolution:
    def __init__(self, nlp=None):
        if nlp is None:
            self.nlp = spacy.load('en', disable=["ner"])
        else:
            self.nlp = nlp
        neuralcoref.add_to_pipe(self.nlp, greedyness=0.5)
        self.code_element_name_util = CodeElementNameUtil()

    @staticmethod
    def replace_with_start_end(start, end, word, text):
        """
        根据起点和终点替换字符串
        :param start:
        :param end:
        :param word:
        :param text:
        :return:
        """
        result = text[:start] + word + text[end:]
        return result

    @staticmethod
    def replace_word_in_string(name_replace_list, text):
        """
        根据起始位置、终止位置、替换单词
        :param name_replace_list:
        :param text:
        :return:
        """
        result = text
        name_replace_list.sort(key=operator.itemgetter(0), reverse=True)
        for name_replace in name_replace_list:
            start_pos, end_pos, old_word, new_word = name_replace
            result = ReferenceResolution.replace_with_start_end(start_pos, end_pos, new_word, result)
        return result

    @staticmethod
    def word_num(text):
        """
        有多少个词
        :param text:
        :return: num int
        """
        return len(str(text).split(" "))

    def lemma_word(self, word):
        """
        词性还原
        :param word:
        :return:
        """
        doc = self.nlp(word)
        result = []
        for token in doc:
            if token.lemma_ == "-PRON-":
                result.append(token.norm_)
            else:
                result.append(token.lemma_)
        return " ".join(result)

    def maybe_contain_api(self, name):
        """
        判断name里面是否可能包含API
        :param name:
        :return: True or False
        """
        if name is None or str(name) == "":
            return False

        uncamelize = self.code_element_name_util.uncamelize(name.strip()).replace("( ", "(")
        if len(name.strip().split(" ")) < len(uncamelize.strip().split(" ")) or self.check_upper_case(name):
            return True
        return False

    def check_pronoun(self, word):
        doc = self.nlp(word)
        for token in doc:
            if token.lemma_ == '-PRON-':
                return True
        # pronoun_set = {
        #     "it", "that", "this", "these", "those",
        #     "which", "one", "ones", "they", "other", "another"
        # }
        # for n in str(word).split(" "):
        #     if n in pronoun_set:
        #         return True
        return False

    @staticmethod
    def replace_ignore_case(text, old, new, ):
        reg = re.compile(re.escape(old), re.IGNORECASE)
        return str(reg.sub(new, text))

    def remove_reference_with_context(self, context, text, input_doc=None):
        """

        :param input_doc:
        :param context: 一个字典，可以传入方法需要的上下文信息，
        package 信息，class信息，method信息，文档归属对象类型
        :param text: 文本
        :return: text，每句话的改动
        """
        # 按照词数目进行过滤
        skip_size = 4
        may_be_class_set = {"this class", "the class"}
        if input_doc is None:
            doc = self.nlp(text)
        else:
            doc = input_doc
        row = doc.text
        result = []
        if doc._.coref_clusters is None:
            return row
        for cluster in doc._.coref_clusters:
            try:
                new_word = cluster.main.text
                if ReferenceResolution.word_num(new_word) >= skip_size:
                    continue
                for men in cluster.mentions:
                    if cluster.main.text == men.text:
                        continue
                    old_word = men.text
                    start_pos = men.start_char
                    end_pos = men.end_char
                    for may_class in may_be_class_set:
                        if self.lemma_word(new_word).find(may_class) >= 0:
                            new_word = context["qn"]
                    if new_word == "":
                        continue
                    # 判断old_word是否已经是API了,已经是了就不替换
                    if self.maybe_contain_api(old_word) or self.lemma_word(old_word) == self.lemma_word(
                            new_word):
                        continue
                    # 判断old_word是否是代词，不是就continue
                    if not self.check_pronoun(old_word):
                        continue
                    # 判断new_word是否是API
                    if not self.maybe_contain_api(new_word):
                        continue
                    result.append((start_pos, end_pos, old_word, new_word))
            except Exception as e:
                print(e)
        after_text = ReferenceResolution.replace_word_in_string(result, row)
        after_text = ReferenceResolution.replace_ignore_case(after_text, "this one", context["qn"])
        after_text = ReferenceResolution.replace_ignore_case(after_text, "this class", context["qn"])
        return after_text, result

    @staticmethod
    def check_upper_case(text):
        """
        判断文本里面是否包含了驼峰式的class
        :param text:
        :return:
        """
        splited = str(text).split(" ")
        for i, s in enumerate(splited):
            if i > 0 and s[0].isupper():
                return True
        return False


# if __name__ == '__main__':
#     r = ReferenceResolution()
#     # print(r.maybe_contain_api("The class is a"))
#     s = """A mutable sequence of characters. This class provides an API compatible with StringBuffer, but with no guarantee of synchronization. This class is designed for use as a drop-in replacement for StringBuffer in places where the string buffer was being used by a single thread (as is generally the case). Where possible, it is recommended that this class be used in preference to StringBuffer as it will be faster under most implementations.
# The principal operations on a StringBuilder are the append and insert methods, which are overloaded so as to accept data of any type. Each effectively converts a given datum to a string and then appends or inserts the characters of that string to the string builder. The append method always adds these characters at the end of the builder; the insert method adds the characters at a specified point.
#
# For example, if z refers to a string builder object whose current contents are "start", then the method call z.append("le") would cause the string builder to contain "startle", whereas z.insert(4, "le") would alter the string builder to contain "starlet".
#
# In general, if sb refers to an instance of a StringBuilder, then sb.append(x) has the same effect as sb.insert(sb.length(), x).
#
# Every string builder has a capacity. As long as the length of the character sequence contained in the string builder does not exceed the capacity, it is not necessary to allocate a new internal buffer. If the internal buffer overflows, it is automatically made larger.
#
# Instances of StringBuilder are not safe for use by multiple threads. If such synchronization is required then it is recommended that StringBuffer be used."""
#     s = """The System class contains several useful class fields and methods. It cannot be instantiated."""
#     context = {"qn": "java.lang.StringBuilder"}
#     tmp = "it"
#     r.lemma_word(tmp)
#     remove_result_text, change_list = r.remove_reference_with_context(context, s)
#     # r.remove_reference(s)
#     print(remove_result_text)



from unittest import TestCase

from sekg.term.fusion import Fusion


class TestTerm(TestCase):

    def test_fusion(self):
        fusion_tool = Fusion()
        terms = {"annotated bind", "low bind", "binds", "ok", "low of bind", "string tc", "Fine", "good", "pretty"}
        synsets = fusion_tool.fuse_by_synonym(terms)
        print(synsets)


class TestFusion(TestCase):
    def test_checkT_synnym(self):
        fusion_tool = Fusion()
        terms = {"RETURN","rule","return","annotated binds","annotated bind", "low bind", "binds", "ok", "low of bind", "string tc", "Fine", "good", "pretty","mode","model"}


        synsets = fusion_tool.fuse_by_synonym(terms)
        print(synsets)
        for t in synsets:
            print(t)

    def test_checkT_synnym_for_time(self):
        fusion_tool = Fusion()

        terms =[]

        for t in range(0,1000):
            terms.append("hhh"+str(t))

        synsets = fusion_tool.fuse_by_synonym(terms)
        print(synsets)
        for t in synsets:
            print(t)


import random
from pathlib import Path
from unittest import TestCase

from sekg.ir.doc.wrapper import MultiFieldDocumentCollection
from sekg.ir.preprocessor.code_text import CodeDocPreprocessor
from test.data.definition import ROOT_DIR


class TestCodeDocPreprocessor(TestCase):
    def test_clean(self):
        data_dir = Path(ROOT_DIR)

        preprossor = CodeDocPreprocessor()
        doc_collection: MultiFieldDocumentCollection = MultiFieldDocumentCollection.load(
            str(data_dir / "JabRef-2.6.v3.sub.dc"))

        chosen_indexs = list(range(0, doc_collection.get_num()))
        random.shuffle(chosen_indexs)
        chosen_indexs = chosen_indexs[:50]
        for index in chosen_indexs:
            doc = doc_collection.get_by_index(index=index)
            text = doc.get_document_text()
            print("old text=", text)
            print("------------")
            clean = preprossor.clean(text=text)
            print("clean text=", " ".join(clean))
            print("------------\n\n")



from unittest import TestCase

from sekg.wiki.wiki_util import WikiDataPropertyTable


class TestWikiDataPropertyTable(TestCase):
    def test_property_id_2_name(self):
        table = WikiDataPropertyTable()
        self.assertEqual("subclass of", table.property_id_2_name("P279"))
        self.assertEqual(None, table.property_id_2_name("xxx"))

    def test_property_name_2_id(self):
        table = WikiDataPropertyTable()
        self.assertEqual("P279", table.property_name_2_id("subclass of"))
        self.assertEqual(None, table.property_name_2_id("xxx"))



from unittest import TestCase

from py2neo import Graph

from sekg.graph.accessor import GraphAccessor
from sekg.graph.metadata_accessor import MetadataGraphAccessor
from sekg.graph.creator import NodeBuilder
from sekg.graph.factory import GraphInstanceFactory
from sekg.graph.index_accessor import IndexGraphAccessor


class TestMetadataAccessor(TestCase):
    test_server_name = "LocalHost"
    test_config_path = "neo4j_config.json"
    test_server_id = 3

    def test_get_max_id_for_node(self):
        factory = GraphInstanceFactory(self.test_config_path)
        graph = factory.create_py2neo_graph_by_server_id(self.test_server_id)
        accessor = MetadataGraphAccessor(graph=graph)
        result = accessor.get_max_id_for_node()
        print(result)

    def test_get_max_id_for_relation(self):
        factory = GraphInstanceFactory(self.test_config_path)
        graph = factory.create_py2neo_graph_by_server_id(self.test_server_id)
        accessor = MetadataGraphAccessor(graph=graph)
        result = accessor.get_max_id_for_relation()
        print(result)

    def test_get_node_num(self):
        factory = GraphInstanceFactory(self.test_config_path)
        graph = factory.create_py2neo_graph_by_server_id(self.test_server_id)
        accessor = MetadataGraphAccessor(graph=graph)
        result_1 = accessor.get_node_num()
        print("all nodes: ", result_1)
        result_2 = accessor.get_node_num("api method")
        print("api method: ", result_2)

    def test_get_relation_num(self):
        factory = GraphInstanceFactory(self.test_config_path)
        graph = factory.create_py2neo_graph_by_server_id(self.test_server_id)
        accessor = MetadataGraphAccessor(graph=graph)
        result = accessor.get_relation_num()
        print(result)

    def test_count_relation_type(self):
        factory = GraphInstanceFactory(self.test_config_path)
        graph = factory.create_py2neo_graph_by_server_id(self.test_server_id)
        accessor = MetadataGraphAccessor(graph=graph)
        result = accessor.count_relation_type()
        print(result)

    def test_expand_node(self):
        factory = GraphInstanceFactory(self.test_config_path)
        graph = factory.create_py2neo_graph_by_server_id(self.test_server_id)
        accessor = MetadataGraphAccessor(graph=graph)
        result = accessor.expand_node(2226, 10)
        print(result)
